/*
    kavr
 
    AVR Hexfile Boot Loader (pronounced "caviar")
 
    Copyright (C) 2009-2011 Frank Edelhaeuser
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include "sp_driver.h"


// macros
#define PACKED          __attribute__((__packed__))
#define NOINLINE        __attribute__((noinline))


// definitions
#define DEBUG		0
#define XON             0x11
#define XOFF            0x13
#define TIMEOUT_LOOPS   (((F_CPU/1000)*TIMEOUT)/42)
#define FIFO_SIZE       1024
#define FIFO_MAX_FILL   (FIFO_SIZE - (FIFO_SIZE / 10))
#define FIFO_MIN_FILL   (FIFO_SIZE - (FIFO_SIZE / 5))
#define FIFO_EOF	0xFFFF


// structures
typedef struct {
    uint8_t     length;
    uint16_t    address;
    uint8_t     type;
    uint8_t     data[255 + 1];
} PACKED hex_record_t;

typedef struct {
  uint16_t in_ptr, out_ptr, count;
  uint8_t  buffer[FIFO_SIZE];
} PACKED fifo_t;


// data
static uint8_t  write_buffer[APP_SECTION_PAGE_SIZE];
static fifo_t	fifo;


// prototypes
static void uart_init(void) NOINLINE;
static void uart_stop(void) NOINLINE;
static void uart_poll(void) NOINLINE;
static void out_xon() NOINLINE;
static void out_xoff() NOINLINE;
static void out_char(uint8_t c) NOINLINE;
static void out_info(uint8_t c, uint32_t value) NOINLINE;
static void fifo_put (uint8_t c) NOINLINE;
static uint8_t fifo_get (void) NOINLINE;
static void flash_write_buffer(uint8_t bank, uint16_t pageAddress) NOINLINE;

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// implementation
int main(void) {
  hex_record_t record;
  uint32_t timeout;
  uint16_t new_page, cur_page, addr, bank, row_cnt;
  uint8_t *rx_ptr, rx_dat, rx_crc, rx_len, rx_val, i;

  // activate 32 MHz oszillator
  OSC.CTRL |= OSC_RC32MEN_bm;
  // wait for oszillator ready
  while(!(OSC.STATUS & OSC_RC32MRDY_bm));
  // lock it
  CCP = CCP_IOREG_gc;
  // use it
  CLK.CTRL = CLK_SCLKSEL_RC32M_gc;

  uart_init();

  // Hello world!
  out_char('K');
  out_char('A');
  out_char('V');
  out_char('R');
  out_char('X');
  out_char('\n');

  // Initialize overall state
  bank     = 0;
  cur_page = 0;
  timeout  = 0;
  row_cnt  = 0;
  fifo.in_ptr  = 0;
  fifo.out_ptr = 0;
  fifo.count   = 0;

  out_xon();

  // (Re-)Initialize line state
  restart_line:
  rx_val = 0;
  rx_len = 0;
  rx_crc = 0;
  rx_ptr = (uint8_t*) &record;

  // Download data records until timeout
  while (timeout < TIMEOUT_LOOPS) {
    timeout++;

    // look for data from serial line
    uart_poll();

    if (fifo.count > 0) {
      // Extract oldest character from receive FIFO
      rx_dat = fifo_get();

      // Convert hex to binary nibble
      rx_dat -= '0';
      if (rx_dat > 9) {
	rx_dat -= 7;
	if ((rx_dat < 10) || (rx_dat > 15))
	  goto restart_line;
      }

      // Valid hex character. Restart timeout.
      timeout = 0;

      // Two nibbles = 1 byte
      rx_val = (rx_val << 4) | rx_dat;
      if ((++rx_len & 1) == 0) {
	rx_crc += rx_val;
	*rx_ptr++ = rx_val;

	if (rx_ptr == (uint8_t*)&record.data[record.length + 1]) {
	  out_xoff();
	  uart_poll();
	  if (DEBUG) out_info('r', row_cnt);
	  row_cnt++;
	  if (rx_crc != 0) {
	    // Checksum error. Abort
	    out_char('?');
	    break;
	  }
	  if (record.type == 0) {
	    // Data record (swap bytes in address value)
	    addr = (record.address << 8) | (record.address >> 8);
	    if (DEBUG) out_info('a', addr);
	    for (i = 0; i < record.length; i++, addr++) {
	      if (DEBUG) out_info('i', i);
	      // Determine write_buffer base for current address
	      new_page = addr & ~(APP_SECTION_PAGE_SIZE - 1);
	      if (new_page != cur_page) {
		if (DEBUG) out_info('n', new_page);
		// Write updated RAM write_buffer into flash
		flash_write_buffer(bank, cur_page);

		// Load write_buffer at new address into RAM write_buffer buffer
		cur_page = new_page;
	      }

	      // Update RAM write_buffer buffer from data record
	      write_buffer[addr & (APP_SECTION_PAGE_SIZE - 1)] = record.data[i];
	    }

	    out_xon();
	    goto restart_line;
	  }
	  if (record.type == 1 || record.type == 4) {
	    // End of file record or bank select. Finish
	    flash_write_buffer(bank, cur_page);
	    out_char('S');
	    break;
	  }
	  if (record.type == 4) {
	    // bank select for devices with more than 64k flash
	    bank = ((uint16_t)record.data[0] << 8) | record.data[1];
	  }
	  out_xon();
	}
      }
    }
  }

  uart_stop();
  RAMPZ = 0;
  asm("jmp 0x0000");                // Run application code
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void fifo_put (uint8_t c) {
  if (DEBUG) out_char(c);
  if (fifo.count == FIFO_SIZE) return;
  fifo.buffer[fifo.in_ptr] = c;
  fifo.count++;
  fifo.in_ptr = (fifo.in_ptr +1) % FIFO_SIZE;
  if (fifo.count == FIFO_MAX_FILL) {
    out_xoff();
  }
}


static uint8_t fifo_get (void) {
  uint8_t c;

  if (fifo.count == 0) return 0;
  c = fifo.buffer[fifo.out_ptr];
  fifo.count--;
  fifo.out_ptr = (fifo.out_ptr +1) % FIFO_SIZE;
  if (fifo.count == FIFO_MIN_FILL) {
    out_xon();
  }
  return c;
} 

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

// UART_PORT  = PORTD
// UART_NAME  = USARTD1
// UART_ISR   = USARTD1_RXC_vect
// UART_TXPIN = 7
// UART_BAUD  = 115200

#define BSCALE	0
#define BSEL	(uint16_t)((F_CPU / ((1 << (BSCALE + 4)) * UART_BAUD)) -1)


static void uart_init(void) {
  // Tx Port Pin output
  UART_PORT.OUT = (1 << (UART_TXPIN));
  UART_PORT.DIR = (1 << (UART_TXPIN));

  // set baud rate
  UART_NAME.BAUDCTRLA = (uint8_t)(BSEL & 0x00FF);
  UART_NAME.BAUDCTRLB = (BSCALE << USART_BSCALE_gp);
  // frame format 8 bits
  UART_NAME.CTRLC = USART_CHSIZE_8BIT_gc;
  // no inetrrupt
  UART_NAME.CTRLA = 0;
  // enable send and receive
  UART_NAME.CTRLB = USART_TXEN_bm | USART_RXEN_bm;
}


static void uart_stop(void) {
  UART_NAME.CTRLB = 0;
}


static void uart_poll(void) {
  while (UART_NAME.STATUS & USART_RXCIF_bm) {
    fifo_put(UART_NAME.DATA);
  }
}


static void out_xon() {
  if (DEBUG) out_char('+');
  out_char(XON);
}


static void out_xoff() {
  if (DEBUG) out_char('-');
  out_char(XOFF);
}


static void out_char(uint8_t c) {
  while (!(UART_NAME.STATUS & USART_DREIF_bm));
  UART_NAME.DATA = c;
}


// for debug purposes only:

static void out_str(char *str) {
  while (*str != '\0') {
    out_char(*str);
    str++;
  }
}


static void out_info(uint8_t c, uint32_t value) {
  char buf[32];

  out_char(c);
  out_char('=');
  out_str(ltoa(value, buf, 10));
  out_char('\n');
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void flash_write_buffer(uint8_t bank, uint16_t pageAddress) {
  uint16_t i;
  uint32_t flashAddress;
  uint8_t  *srcPtr, lowByte, highByte;

  if (DEBUG) out_info('P', pageAddress);

  // Copy RAM addr buffer into SPM addr buffer
  if (DEBUG) out_char('l');
  srcPtr = write_buffer;
  for (i = 0; i < APP_SECTION_PAGE_SIZE; i += 2) {
    lowByte  = *srcPtr++;
    highByte = *srcPtr++;
    SP_LoadFlashWord(i, ((uint16_t)highByte << 8) | lowByte);
    while (NVM.STATUS & NVM_NVMBUSY_bm) uart_poll();

    if (DEBUG) out_char('d');
  }
  if (DEBUG) out_char('\n');

  flashAddress = APP_SECTION_START + ((uint32_t)bank << 16) + pageAddress;
  if (DEBUG) out_info('F', flashAddress);

  // Erase and program
  if (DEBUG) out_char('w');
  SP_EraseWriteApplicationPage(flashAddress);

  // clear write_buffer
  srcPtr = write_buffer;
  for (i = 0; i < APP_SECTION_PAGE_SIZE; i++) {
    *srcPtr++ = 0xFF;
  }

  // wait for NVM ready
  while (NVM.STATUS & NVM_NVMBUSY_bm) uart_poll();

  // clear buffer
  if (DEBUG) out_char('c');
  srcPtr = write_buffer;
  for (i = 0; i < APP_SECTION_PAGE_SIZE; i++) *srcPtr = 0x00;
  if (DEBUG) out_char('\n');
}
